import re


def camelize(string: str) -> str:
    """Transform snake case into camelcase"""

    string = re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), string)
    return string[0].lower() + string[1:]



import os

import pytest
import yaml

from kuku.types import Context
from kuku.values import resolve

HERE = os.path.abspath(os.path.dirname(__file__))
VALUES_STAGING_FILE = os.path.join(HERE, "fixtures/values/values-staging.yaml")
VALUES_PRODUCTION_FILE = os.path.join(HERE, "fixtures/values/values-production.yaml")


@pytest.mark.parametrize(
    "values, expected",
    [
        (["k=v"], [{"k": "v"}]),
        (["k1=v1,k2=v2"], [{"k1": "v1", "k2": "v2"}]),
        (["k1=v1,k2=v2", "k3=v3"], [{"k1": "v1", "k2": "v2"}, {"k3": "v3"}]),
        (["k="], [{"k": ""}]),
    ],
)
def test_valid_values(values, expected):
    assert resolve(values, []) == Context(*expected)


@pytest.mark.parametrize(
    "values, expected",
    [
        (["a.b=v1", "a.c=v2"], [{"a": {"c": "v2"}}, {"a": {"b": "v1"}}]),
        (
            ["a.b.c=v1", "a.b.d=v2"],
            [{"a": {"b": {"d": "v2"}}}, {"a": {"b": {"c": "v1"}}}],
        ),
    ],
)
def test_nested_dicts(values, expected):
    assert resolve(values, []) == Context(*expected)


@pytest.mark.parametrize(
    "values, expected",
    [
        (["a.0.b=v1", "a.0.c=v2"], [{"a": [{"c": "v2"}]}, {"a": [{"b": "v1"}]}]),
        # (["a.0.b=v1", "a.1.c=v2"], [{"a": [{"c": "v2"}]}, {"a": [{"b": "v1"}]}]),
    ],
)
def test_nested_lists(values, expected):
    assert resolve(values, []) == Context(*expected)


@pytest.mark.parametrize(
    "values", [["k"], ["=v"], ["="], ["k=1,=2"], ["a.b.c=1,=2"], ["a.0.c=1,=2"]]
)
def test_resolve_invalid_values(values):
    with pytest.raises(ValueError):
        resolve(values, [])


def test_cli_values_value_files():
    values = resolve([], [VALUES_STAGING_FILE])
    with open(VALUES_STAGING_FILE) as fd:
        assert values == yaml.load(fd)

    # putting the same file twice should not change the resulting values
    assert resolve([], [VALUES_STAGING_FILE, VALUES_STAGING_FILE]) == values

    # cli values override values from files
    assert resolve(["env=production"], [VALUES_STAGING_FILE])["env"] == "production"

    # multiple values override (last wins)
    assert resolve(["env=1", "env=2"], [VALUES_STAGING_FILE])["env"] == "2"

    # nested override
    assert (
        resolve(["nodeSelector.node-label=pool-2"], [VALUES_STAGING_FILE])[
            "nodeSelector"
        ]["node-label"]
        == "pool-2"
    )


def test_value_files_override():
    values = resolve([], [VALUES_STAGING_FILE])
    assert values["nodeSelector"]["node-label"] == "pool-1"

    values = resolve([], [VALUES_STAGING_FILE, VALUES_PRODUCTION_FILE])
    assert values["nodeSelector"]["node-label"] == "pool-2"



from typing import Dict, List, Union, Any

from kuku.types import IgnoredListItem

DictOrList = Union[Dict, List]


def unroll_key(key, value) -> Dict[str, Any]:
    """Unroll a dot notation key (i.e: k1.k2.0.k3) to a dictionary structure and set it's last leaf to `value`"""

    if "." not in key:
        return {key: value}

    def walk(path):
        # last item -> set value
        if not path:
            return value

        item = path[0]
        try:
            index = int(item)
            if str(index) == item:
                res = []
                for i in range(index + 1):
                    res.append(IgnoredListItem if i != index else walk(path[1:]))
                return res
        except ValueError:
            pass

        return {item: walk(path[1:])}

    return walk(key.split("."))



from kubernetes import client


def template(context):
    return client.V1Service(
        api_version="v1",
        kind="Service",
        metadata=client.V1ObjectMeta(name=context["name"]),
        spec=client.V1ServiceSpec(
            type="NodePort",
            ports=[
                {"port": context["externalPort"], "targetPort": context["internalPort"]}
            ],
            selector={"app": context["name"]},
        ),
    )



import sys

from docopt import docopt
from kubernetes import config

from kuku.__version__ import __version__
from kuku.dump import dump
from kuku.render import render
from kuku.templates import find
from kuku.values import resolve


def cli():
    """kuku: Kubernetes templating tool.
    
    Usage:
      kuku apply [-v] [-f <FILE>]... [-s <key=value>]... <TEMPLATES_DIR>
      kuku delete [-v] [-f <FILE>]... [-s <key=value>]... <TEMPLATES_DIR>
      kuku render [-v] [-f <FILE>]... [-s <key=value>]... <TEMPLATES_DIR>
      kuku (-h | --help)
      kuku --version

    Options:
      -h --help                     Show this screen.
      -v --verbose                  Dump debug info to stderr.
      --version                     Show version.
      -s KEY=VALUE --set KEY=VALUE  Set values on the command line (accepts multiple options or separate values with commas: Ex: -s key1=val1,key2=val2).
      -f FILE --file FILE           Specify values in a YAML file (accepts multiple options).

    Notes:
      Resolution of values: --set overrides values in --file by merging. The last value wins.
    """

    # Parse cli arguments from docstring
    arguments = docopt(str(cli.__doc__), version=__version__)

    # Load k8s config -- needed to the correct API versions of k8s
    config.load_kube_config()

    # Find all templates
    try:
        templates = find(arguments["<TEMPLATES_DIR>"])
    except ValueError as e:
        print(e)
        exit(1)
        raise

    try:
        # Resolve values
        context = resolve(arguments["--set"], arguments["--file"])
    except ValueError as e:
        print(e)
        exit(1)
        raise

    # Render templates with resolved context
    rendering = render(context, templates)

    output = dump(rendering)
    if arguments["render"]:
        # print yaml
        print(output)

    if arguments["apply"]:
        print("Not implemented yet. Use kuku render ... | kubectl apply -f-")
        exit(1)

    if arguments["delete"]:
        print("Not implemented yet. Use kuku render ... | kubectl delete -f-")
        exit(1)

    if arguments["--verbose"]:
        print(output, file=sys.stderr)


if __name__ == "__main__":
    cli()



from kubernetes import client


def template(context):
    return client.V1Namespace(
        api_version="v1",
        kind="Namespace",
        metadata=client.V1ObjectMeta(name=context["name"]),
    )



import yaml

# don't dump with aliases
noalias_dumper = yaml.dumper.SafeDumper
noalias_dumper.ignore_aliases = lambda self, data: True  # type: ignore


def str_presenter(yaml_dumper, data):
    """Use block style for multiline strings"""
    if len(data.splitlines()) > 1:  # check for multiline string
        return yaml_dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return yaml_dumper.represent_scalar("tag:yaml.org,2002:str", data)


noalias_dumper.add_representer(str, str_presenter)


def dumper(data):
    return yaml.dump(data, default_flow_style=False, Dumper=noalias_dumper)



from typing import List

import yaml

from kuku.types import Context
from kuku.utils.dict import unroll_key


def resolve(values: List[str], value_files: List[str]) -> Context:
    """Resolve values from cli and value files"""

    chain = []
    resolved_key_values = []

    # arguments from the CLI have priority
    for key_values in reversed(values):
        if "," in key_values:
            extended_key_values = key_values.split(",")
        else:
            extended_key_values = [key_values]

        for key_value in extended_key_values:
            format_error = "Invalid key=value format for: {}".format(key_value)
            if "=" not in key_value:
                raise ValueError(format_error)

            key, value = key_value.split("=")
            if not key:
                raise ValueError(format_error)

            resolved_key_values.append(unroll_key(key, value))

    chain.extend(resolved_key_values)

    for value_file in reversed(value_files):
        with open(value_file) as fd:
            chain.append(yaml.load(fd))

    return Context(*chain)



import os

import pytest

from kuku.templates import find

HERE = os.path.abspath(os.path.dirname(__file__))


def test_find_templates():
    templates = find(os.path.join(HERE, "fixtures/templates"))
    assert len(templates) == 2
    assert all([t.endswith(".py") for t, v in templates.items()])

    with pytest.raises(ValueError):
        find(os.path.join(HERE, "fixtures/values"))



from typing import Dict, Any, Callable, List
from collections import ChainMap

# Context is used to pass variables
Context = ChainMap

# template file path -> template function
Templates = Dict[str, Callable]

# template file path -> List of K8S objects
Rendering = Dict[str, List[Any]]


class IgnoredListItem:
    """Placeholder list item to be ignored for deep merges of lists"""

    ...



from types import MethodType

from kuku.types import Rendering
from kuku.utils.str import camelize
from kuku.utils.yaml import dumper

OBJECT_SEPARATOR = "\n---\n"


def _camelized_to_dict(self):
    """Override the default k8s object to_dict to camelize it's keys"""

    result = {}

    for attr, _ in self.swagger_types.items():
        value = getattr(self, attr)

        # we need the attribute camelized for YAML, also ignore the _ prefixes.
        attr = camelize(attr.lstrip("_"))

        if isinstance(value, list):
            result[attr] = list(
                map(
                    lambda x: _camelized_to_dict(x) if hasattr(x, "to_dict") else x,
                    value,
                )
            )
        elif hasattr(value, "to_dict"):
            result[attr] = _camelized_to_dict(value)
        elif isinstance(value, dict):
            result[attr] = dict(
                map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict")
                    else item,
                    value.items(),
                )
            )
        else:
            # ignore None values - we don't need the for the output
            if value is not None:
                result[attr] = value

    return result


def dump(rendering: Rendering) -> str:
    """Dump all templates rendering into a yaml string"""

    full_output = []
    for template_path, k8s_objects in rendering.items():
        template_output = []
        template_header = "# Source: {}\n".format(template_path)
        for k8s_object in k8s_objects:
            # Override the default to_dict method so we can update the k8s keys
            k8s_object.to_dict = MethodType(_camelized_to_dict, k8s_object)
            k8s_object = k8s_object.to_dict()

            template_output.append(dumper(k8s_object) + "\n")
        full_output.append(template_header + OBJECT_SEPARATOR.join(template_output))
    return OBJECT_SEPARATOR.join(full_output)



#!/usr/bin/env python

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = "kuku"
DESCRIPTION = "Kubernetes templating tool."
URL = "https://github.com/xarg/kuku"
EMAIL = "alex@plugaru.org"
AUTHOR = "Alex Plugaru"
REQUIRES_PYTHON = ">=3.5.0"

REQUIRED = ["docopt", "PyYAML", "requests", "kubernetes"]

EXTRAS = {
    "dev": ["twine", "wheel", "pytest", "pytest-cov", "black", "mypy", "pre-commit"]
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, NAME, "__version__.py")) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=("tests",)),
    py_modules=["kuku"],
    entry_points={"console_scripts": ["kuku=kuku.cli:cli"]},
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="Apache",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



from copy import deepcopy

from kuku.types import Context, Templates, Rendering


def render(context: Context, templates: Templates) -> Rendering:
    """Given a `context` of values and a list of `templates` render them to k8s objects"""

    rendering = {}
    for template_path, template_func in templates.items():
        # pass a copy of the context to the template function and get it's k8s objects
        k8s_objects = template_func(deepcopy(context))
        if not isinstance(k8s_objects, (list, tuple)):
            k8s_objects = [k8s_objects]
        rendering[template_path] = k8s_objects

    return rendering



import os

import pytest

from kuku.render import render
from kuku.templates import find
from kuku.types import Context

HERE = os.path.abspath(os.path.dirname(__file__))


def test_render():
    templates = find(os.path.join(HERE, "fixtures/templates"))

    with pytest.raises(KeyError):
        render(Context(), templates)

    rendering = render(
        Context({"name": "test", "internalPort": 80, "externalPort": 80}), templates
    )

    assert len(rendering) == len(templates)



import os
import pkgutil

from kuku.types import Templates

TEMPLATE_FUNCTION_NAME = "template"


def find(templates_dir: str) -> Templates:
    """Given a templates directory and import all python templates"""

    templates = {}

    for module_info in pkgutil.iter_modules([templates_dir]):
        module = module_info.module_finder.find_module(module_info.name).load_module()
        template_path = os.path.join(templates_dir, module_info.name + ".py")
        # only consider python modules with a template function
        template_func = getattr(module, TEMPLATE_FUNCTION_NAME, None)
        if template_func:
            templates[template_path] = template_func

    if not templates:
        raise ValueError(
            "No kuku python templates files found in {}".format(templates_dir)
        )

    return templates
